import os
import sys
sys.path.insert(0, os.path.abspath(".."))

project = "Python LED Strip Bluetooth Controller API"
copyright = "2024, Douglas Paz"
author = "Douglas Paz"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
]

templates_path = ["_templates"]
exclude_patterns = []

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



"""
controllers

supported controllers:
- MELK: MELKController

inspired by:
- https://github.com/dave-code-ruiz/elkbledom/blob/main/custom_components/elkbledom/elkbledom.py
"""

from typing import (
    Any,
    List,
    Optional,
)
import datetime
from .typing import Command

COMMAND_PREFIX = "command_"

class Controller:
    """
    base controller class
    """
    _char_specifier: Optional[str] = None

    @property
    def char_specifier(self) -> str:
        """
        char specifier
        """
        assert self._char_specifier
        return self._char_specifier

    def init_commands(self) -> List[Command]:
        """
        init commands
        """
        return []

    def build_command(self, *args: int) -> Command:
        """
        build generic command
        """
        raise NotImplementedError()

    def __getattr__(self, name: str) -> Any:
        if not name.startswith(COMMAND_PREFIX):
            return super().__getattr__(name)
        act = name.removeprefix(COMMAND_PREFIX)
        command_fn = getattr(self, f"_{act}", None)
        if command_fn:
            return command_fn
        def command_wrapper():
            return getattr(self, f"_{COMMAND_PREFIX}{act}")
        return command_wrapper

class MELKController(Controller):  # pylint: disable=R0903
    """
    MELK controller devices

    Implements:

    - BTLedStrip.exec.turn_on()
    - BTLedStrip.exec.turn_off()
    - BTLedStrip.exec.brightness(percentage: int)
    - BTLedStrip.exec.color(red: float, green: float, blue: float)
    - BTLedStrip.exec.white(cold: int)
    - BTLedStrip.exec.white_cold(brightness: int)
    - BTLedStrip.exec.effect(number: int)
    - BTLedStrip.exec.effect_speed(velocity: int)
    """
    _char_specifier = "0000fff3-0000-1000-8000-00805f9b34fb"
    _command_turn_on: Command = [0x7e, 0x00, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0xef]
    _command_turn_off: Command = [0x7e, 0x00, 0x04, 0x00, 0x00, 0x00, 0xff, 0x00, 0xef]

    def init_commands(self) -> List[Command]:
        date = datetime.date.today()
        now = datetime.datetime.now()
        _, _, day_of_week = date.isocalendar()
        return [[0x7e, 0x07, 0x83],
                [0x7e, 0x04, 0x04],
                [0x7e, 0x00, 0x83, int(now.strftime('%H')), int(now.strftime('%M')),
                 int(now.strftime('%S')), day_of_week, 0x00, 0xef]]

    def build_command(self,  # pylint: disable=W0221,R0913
                      arg_1: int,
                      arg_2: int,
                      arg_3: int,
                      arg_4: int,
                      arg_5: int,
                      arg_6: int) -> Command:
        return [0x7e, arg_1, arg_2, arg_3, arg_4, arg_5, arg_6, 0x00, 0xef]

    def _brightness(self, percentage: int = 0) -> Command:
        """
        set brightness on color mode
        """
        return [0x7e, 0x04, 0x01, percentage, 0x00, 0x00, 0x00, 0x00, 0xef]

    def _color(self, red: float = 0, green: float = 0, blue: float = 0) -> Command:
        """
        mode color
        """
        r = int(red * 255 / 100)
        g = int(green * 255 / 100)
        b = int(blue * 255 / 100)
        return [0x7e, 0x00, 0x05, 0x03, r, g, b, 0x00, 0xef]

    def _white(self, cold: int = 0) -> Command:
        """
        mode white with temperature
        """
        return [0x7e, 0x00, 0x05, 0x02, 0x00, cold, 0x00, 0xef]

    def _white_cold(self, brightness: int = 0) -> Command:
        """
        mode white cold with brightness
        """
        return [0x7e, 0x00, 0x05, 0x01, brightness, 0x00, 0x00, 0xef]

    def _effect(self, number: int) -> Command:
        """
        mode effect
        """
        return [0x7e, 0x04, 0x03, number, 0x00, 0x00, 0x00, 0x00, 0xef]

    def _effect_speed(self, velocity: int = 0):
        """
        set effect speed
        """
        v = int(velocity)
        return [0x7e, 0x04, 0x02, v, 0x00, 0x00, 0x00, 0x00, 0xef]



"""
Dev Command Terminal

Debug a controller sending programmatically commands via Bluetooth.

usage: python -m btledstrip.dev -h
"""
import argparse
import logging
import asyncio
from typing import Type
from btledstrip import (
    BTLedStrip,
    Controller,
    MELKController
)

logger = logging.getLogger(__name__)

CONTROLLERS_CLASS = {
    "melk": MELKController
}

class ControllerClassAction(argparse.Action):
    """
    argparse controller class action
    """
    def __call__(self, _, namespace, values, option_string=None) -> None:
        setattr(namespace, self.dest, CONTROLLERS_CLASS[values])

async def main(controller_class: Type[Controller], mac_address: str):
    """
    main
    """
    controller = controller_class()
    async with BTLedStrip(controller, mac_address) as led_strip:
        command_input = input("> ")
        while command_input != "exit":
            command_args = list(map(int, command_input.split()))
            try:
                await led_strip.exec_command(*command_args)
            except Exception as e:  # pylint: disable=W0718
                logger.error(e)
            command_input = input("> ")

parser = argparse.ArgumentParser(prog="btledstrip Dev Command Terminal",
                                 description="Debug a controller sending programmatically commands "
                                             "via Bluetooth")
parser.add_argument("--debug",
                    default=False,
                    action=argparse.BooleanOptionalAction)
parser.add_argument("controller_class",
                    choices=CONTROLLERS_CLASS,
                    action=ControllerClassAction)
parser.add_argument("mac_address")

if __name__ == "__main__":
    args = parser.parse_args()
    logging.basicConfig(level=(logging.DEBUG if args.debug else logging.INFO))
    asyncio.run(main(args.controller_class, args.mac_address))



"""
melk demonstration
"""
import logging
import argparse
import asyncio
from btledstrip import (
    BTLedStrip,
    MELKController,
)

async def main(mac_address: str) -> None:
    """
    run demonstration
    """
    controller = MELKController()
    async with BTLedStrip(controller, mac_address) as led_strip:
        await led_strip.exec.turn_on()
        while True:
            await led_strip.exec.brightness(percentage=100)
            await led_strip.exec.color(red=100)
            await asyncio.sleep(1)
            await led_strip.exec.color(green=100)
            await asyncio.sleep(1)
            await led_strip.exec.color(blue=100)
            await asyncio.sleep(1)
            p = 0
            await led_strip.exec.brightness(percentage=0)
            await led_strip.exec.color(green=100, blue=100)
            while p <= 100:
                await led_strip.exec.brightness(percentage=p)
                await asyncio.sleep(.05)
                p += 1
            await asyncio.sleep(1)
            await led_strip.exec.white()
            await asyncio.sleep(1)
            c = 0
            while c <= 100:
                await led_strip.exec.white(cold=c)
                await asyncio.sleep(.05)
                c += 1
            await asyncio.sleep(1)
            await led_strip.exec.white_cold(brightness=100)
            b = 100
            while b >= 0:
                await led_strip.exec.white_cold(brightness=b)
                await asyncio.sleep(.05)
                b -= 1
            await asyncio.sleep(1)
            await led_strip.exec.effect_speed(velocity=50)
            n = 0
            while n <= 255:
                await led_strip.exec.effect(number=n)
                await asyncio.sleep(1)
                n += 1

parser = argparse.ArgumentParser(prog='btledstrip MELK Demonstration')
parser.add_argument('mac_address')

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    args = parser.parse_args()
    asyncio.run(main(mac_address=args.mac_address))



"""
btledstrip module
"""
import logging
import asyncio
from typing import (
    Callable,
    Awaitable,
    Optional,
    Any,
)
from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from .controllers import (
    Controller,
    MELKController,
    COMMAND_PREFIX,
)
from .typing import Command

logger = logging.getLogger(__name__)

class BTLedStripExec:  # pylint: disable=R0903
    """
    build BTLedStrip.exec
    """
    def __init__(self, bt_led_strip: "BTLedStrip") -> None:
        self._bt_led_strip = bt_led_strip

    def __getattr__(self, name: str) -> Callable[..., Awaitable[None]]:
        command_fn = getattr(self._bt_led_strip.controller, f"{COMMAND_PREFIX}{name}")
        assert command_fn
        async def command_wrapper(**kwargs):
            command = command_fn(**kwargs)
            await self._bt_led_strip.send_command(command)
            logger.info("exec %s %s successfully executed", name, kwargs)
        return command_wrapper

class BTLedStrip:
    """
    Create an object by passing the controller and its settings along with the bluethooth MAC
    address to control the LED strip.
    """
    exec: BTLedStripExec

    def __init__(self,
                 controller: Controller,
                 mac_address: str) -> None:
        self._controller = controller
        self._bt_client: Optional[BleakClient] = None
        self._characteristic = None
        self._mac_address = mac_address
        self.exec = BTLedStripExec(self)

    async def __aenter__(self) -> "BTLedStrip":
        self._bt_client = BleakClient(self._mac_address)
        await self.bt_client.__aenter__()
        for command in self._controller.init_commands():
            logger.info("%s send init command %s", self, command)
            await self.bt_client.write_gatt_char(
                self.characteristic,
                bytearray(command),
                False
            )
            await asyncio.sleep(1)
        logger.info("BTLedStrip context created %s", self)
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.bt_client.__aexit__(*args)
        self._bt_client = None
        self._characteristic = None
        logger.info("BTLedStrip context destroyed %s", self)

    @property
    def controller(self) -> Controller:
        """
        controller
        """
        return self._controller

    @property
    def bt_client(self) -> BleakClient:
        """
        bleak client
        """
        assert self._bt_client
        return self._bt_client

    @property
    def characteristic(self) -> BleakGATTCharacteristic:
        """
        write characteristic
        """
        if not self._characteristic:
            self._characteristic = self._bt_client.services.get_characteristic(
                self._controller.char_specifier)
        return self._characteristic

    async def send_command(self, command: Command):
        """
        send command via bluetooth
        """
        logger.debug("request %s write command: %s", self.bt_client, command)
        await self.bt_client.write_gatt_char(
            self.characteristic,
            bytearray(command),
            False
        )

    async def exec_command(self, *args: int) -> None:
        """
        exec command
        """
        command = self._controller.build_command(*args)
        await self.send_command(command)
        logger.info("command successfully executed: %s", args)

__all__ = [
    "BTLedStrip",
    "Controller",
    "MELKController",
    "Command",
]

VERSION = "0.1.0"
